import enum
import re
from typing import Union

from pydispix.canvas import Pixel


class Color(enum.Enum):
    """A set of common colour codes."""
    BLACK = '000000'
    RED = 'FF0000'
    GREEN = '00FF00'
    BLUE = '0000FF'
    YELLOW = 'FFFF00'
    PINK = MAGENTA = 'FF00FF'
    CYAN = LIGHT_BLUE = '00FFFF'
    WHITE = 'FFFFFF'

    BLURPLE = DISCORD_BLURPLE = '5B55F2'
    DISCORD_RED = 'ED4245'
    DISCORD_GREEN = '57F287'
    DISCORD_YELLOW = 'FEE752'
    DISCORD_PINK = 'EB458E'
    DISCORD_BLACK = '23272A'


def parse_color(value: Union[int, str, tuple[int, int, int], Color]) -> str:
    """Parse a colour to a hex string.
    Accepts integers, strings and instances of the Colour enum.
    """
    if isinstance(value, int):
        if value >= 0 and value <= 0xFFFFFF:
            return f'{value:0>6x}'
    elif isinstance(value, str):
        neat_value = value.lstrip('#').upper()
        if re.match('[0-9A-F]{6}', neat_value):
            return neat_value
        if value.upper() in Colour.__members__:
            return Colour.__members__[value.upper()].value
    elif isinstance(value, Color):
        return value.value
    elif isinstance(value, Pixel):
        # Remove leading "#".
        return str(value).removeprefix("#")
    elif isinstance(value, tuple):
        if len(value) == 3:
            for col_byte in value:
                if not isinstance(col_byte, int):
                    raise TypeError(f"Expected tuple of 3 integers, found {col_byte.__class__.__name__}")
                if col_byte < 0 or col_byte > 255:
                    raise ValueError(f"Colors in rgb tuple must follow 0 <= x <= 255, got {col_byte}")

        return f"{value[0]:0>2x}{value[1]:0>2x}{value[2]:0>2x}"

    raise ValueError(f'Invalid colour "{value}".')


Colour = Color
parse_colour = parse_color



import sys
import time
import logging
from collections import defaultdict
import requests
from requests.models import CaseInsensitiveDict

logger = logging.getLogger('pydispix')


class RateLimitedEndpoint:
    def __init__(self, default_delay: int = 0):
        self.requests_limit = None          # Total number of requests before reset time wait
        self.remaining_requests = 1         # Remaining number of requests before reset time wait
        self.reset_time = 0                 # How much time to wait once we hit reset
        self.cooldown_time = 0              # Some endpoints force longer cooldown times
        self.default_delay = default_delay  # If no other limit is found, how long should we wait
        self.anti_spam_delay = 0            # This is hit when multiple tokens are used

    def update_from_headers(self, headers: CaseInsensitiveDict[str]):
        self.remaining_requests = int(headers.get('requests-remaining', 1))
        self.reset_time = int(headers.get('requests-reset', 0))
        self.cooldown_time = int(headers.get('cooldown-reset', 0))
        self.anti_spam_delay = int(headers.get('retry-after', 0))
        if "requests-limit" in headers:
            self.requests_limit = int(headers["requests-limit"])

    def get_wait_time(self):
        if self.anti_spam_delay != 0:
            return self.anti_spam_delay
        if self.cooldown_time != 0:
            return self.cooldown_time
        if self.remaining_requests == 0:
            return self.reset_time

        return self.default_delay

    def sleep(self, seconds: int, *, show_progress: bool = False):
        # Progress bars shouldn't appear if we're waiting less than 5 seconds
        # it tends to be spammy and doesn't really provide much value
        if not show_progress or seconds < 5:
            return time.sleep(seconds)

        toolbar_width = 40

        # Setup toolbar
        sys.stdout.write(f"[{' ' * toolbar_width}]")
        sys.stdout.flush()
        sys.stdout.write("\b" * (toolbar_width + 1))  # return to start of line, after '['

        for i in range(toolbar_width):
            time.sleep(seconds / toolbar_width)
            sys.stdout.write("#")
            sys.stdout.flush()
        sys.stdout.write("]\n")  # this ends the progress bar

    def wait(self, *, show_progress: bool = False):
        if self.anti_spam_delay != 0:
            logger.warning(f"Sleeping for {self.anti_spam_delay}s, anti-spam cooldown triggered!")
            return self.sleep(self.anti_spam_delay, show_progress=show_progress)
        if self.cooldown_time != 0:
            logger.info(f"Sleeping {self.cooldown_time}s, on cooldown.")
            return self.sleep(self.cooldown_time, show_progress=show_progress)
        if self.remaining_requests == 0:
            logger.info(f"Sleeping {self.reset_time}s, on reset.")
            return self.sleep(self.reset_time, show_progress=show_progress)

        logger.debug(f"Sleeping default delay ({self.default_delay}), {self.remaining_requests} requests remaining.")
        return self.sleep(self.default_delay, show_progress=show_progress)


class RateLimiter:
    def __init__(self):
        self.rate_limits = defaultdict(RateLimitedEndpoint)

    def update_from_headers(self, endpoint: str, headers: CaseInsensitiveDict[str]):
        limiter = self.rate_limits[endpoint]
        limiter.update_from_headers(headers)

    def wait(self, endpoint: str, show_progress: bool = False):
        limiter = self.rate_limits[endpoint]
        limiter.wait(show_progress=show_progress)


class RateLimitBreached(Exception):
    def __init__(self, *args, response: requests.Response, **kwargs):
        super().__init__(*args, **kwargs)

        # Get time limits from headers with RateLimitedEndpoint
        temp_rate_limit = RateLimitedEndpoint()
        temp_rate_limit.update_from_headers(response.headers)

        self.requests_limit = temp_rate_limit.requests_limit
        self.reset_time = temp_rate_limit.reset_time
        self.remaining_requests = temp_rate_limit.remaining_requests
        self.cooldown_time = temp_rate_limit.cooldown_time

        # Store the expected wait and the original response which trigerred this exception
        self.expected_wait_time = temp_rate_limit.get_wait_time()
        self.response = response

    def __str__(self):
        s = super().__str__()
        s += f"\nresponse={self.response.content}"
        if self.expected_wait_time != 0:
            s += f"\nexpected_wait_time={self.expected_wait_time}"

        return s



from pydispix.autodraw import AutoDrawer  # noqa
from pydispix.canvas import Canvas, Pixel  # noqa
from pydispix.client import Client  # noqa
from pydispix.color import Color, Colour, parse_color, parse_colour  # noqa
from pydispix.log import setup_logging
from pydispix.ratelimits import RateLimitedEndpoint, RateLimiter  # noqa

setup_logging()



import os
import requests
import logging
from collections import namedtuple
from typing import Union, Optional

from pydispix.ratelimits import RateLimiter, RateLimitBreached
from pydispix.canvas import Canvas, Pixel
from pydispix.color import Color, parse_color

logger = logging.getLogger('pydispix')
Dimensions = namedtuple('Dimensions', ('width', 'height'))


class Client:
    """HTTP client to the pixel API."""
    def __init__(self, token: Optional[str] = None, base_url: str = "https://pixels.pythondiscord.com/"):
        if token is None:
            try:
                token = os.environ['TOKEN']
            except KeyError:
                raise RuntimeError("Unable to load token, 'TOKEN' environmental variable not found.")

        if not base_url.endswith("/"):
            base_url = base_url + "/"

        self.token = token
        self.base_url = base_url
        self.headers = {
            "Authorization": "Bearer " + token,
            "User-Agent": "ItsDrike pydispix",
        }
        self.rate_limiter = RateLimiter()
        self.width, self.height = self.size = self.get_dimensions()

    def make_raw_request(
        self, method: str, endpoint_url: str, *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> requests.Response:
        """
        This method is here purely to make an HTTP request and update the rate limiter.
        Even though this will update the rate limtis, it will not wait for them.
        """
        logger.debug(f"Request: {method} on {endpoint_url} {data=} {params=}.")
        response = requests.request(
            method, self.base_url + endpoint_url,
            json=data,
            params=params,
            headers=self.headers
        )
        self.rate_limiter.update_from_headers(endpoint_url, response.headers)

        if response.status_code == 429:
            logger.debug(f"Request: {method} on {endpoint_url} {data=} {params=} has failed due to rate limitation.")
            raise RateLimitBreached(
                "Request didn't succeed because it was made during a rate-limit phase.",
                response=response
            )
        if response.status_code == 401:
            logger.error("Request failed with 401 (Forbidden) code. This means your API token is most likely invalid.")
            raise requests.HTTPError(f"Received code {response.status_code} - FORBIDDEN: Is your API token correct?")

        if response.status_code != 200:
            raise requests.HTTPError(f"Received code {response.status_code}", response=response)

        return response

    def make_request(
        self,
        method: str,
        endpoint_url: str,
        *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        parse_json: bool = True,
        ratelimit_after: bool = False,
        show_progress: bool = False,
    ) -> Union[bytes, dict]:
        """
        Make a `method` request to given `endpoint_url`, while respecting the API rate limits.
        You can optionally pass JSON `data` and `parameters`.
        You can set `show_progress` to draw a progress bar while waiting for the rate limtis.

        For some requests, you may want to wait for the rate limit timeout *after* the request
        was made. This is needed because with some requests we want the most recent data we
        can get, after waiting out the limit (for example with get_canvas), but with others
        where we want to make our request as soon as possible, and only then wait for rate limits.
        (for example with set_pixel, we don't know how the canvas may have changed by the
        time we have finished waiting)
        """
        while True:
            if not ratelimit_after:
                self.rate_limiter.wait(endpoint_url, show_progress=show_progress)

            try:
                response = self.make_raw_request(method, endpoint_url, data=data, params=params)
            except RateLimitBreached:
                # This can happen with first request when we're rate-limiting afterwards
                # Or when 2 machines are using the same token. When this occurs we continue
                # and make the request again.
                logger.warning("Hit rate limit, repeating request")

                # Wait for the rate limit if we're limiting afterwards
                if ratelimit_after:
                    self.rate_limiter.wait(endpoint_url, show_progress=show_progress)

                continue

            if ratelimit_after:
                self.rate_limiter.wait(endpoint_url, show_progress=show_progress)

            if parse_json:
                return response.json()
            else:
                return response.content

    def get_dimensions(self) -> Dimensions:
        """Make a request to obtain the canvas dimensions"""
        data = self.make_request("GET", "get_size")
        return Dimensions(width=data["width"], height=data["height"])

    def get_canvas(self, show_progress: bool = False) -> Canvas:
        """Fetch the whole canvas and return it in a `Canvas` object."""
        data = self.make_request("GET", "get_pixels", parse_json=False, show_progress=show_progress)
        return Canvas(self.size, data)

    def get_pixel(self, x: int, y: int, show_progress: bool = False) -> Pixel:
        """Fetch rgb data about a specific pixel"""
        data = self.make_request("GET", "get_pixel", params={"x": x, "y": y}, show_progress=show_progress)
        hex_color = data["rgb"]
        return Pixel.from_hex(hex_color)

    def put_pixel(
        self,
        x: int, y: int,
        color: Union[int, str, tuple[int, int, int], Color],
        show_progress: bool = False,
        retry_on_limit: bool = False,
    ) -> str:
        """Draw a pixel and return a message."""
        # Wait for ratelimits *after* making request, not before. This makes
        # sense because we don't know how the canvas may have changed by the
        # time we have finished waiting, whereas for GET endpoints, we want to
        # return the information as soon as it is given.
        data = self.make_request(
            'POST', 'set_pixel',
            data={
                'x': x,
                'y': y,
                'rgb': parse_color(color)
            },
            ratelimit_after=True,
            show_progress=show_progress
        )

        logger.info('Success: {message}'.format(**data))
        return data['message']
